#!/usr/bin/python
"""
Foal - Familiar Open Application Launcher

Copyright (C) 2001 James ''Wez''Weatherall

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Todo's

1.  Fix always-on-top code
2.  Some more standard icons...
3.  Integration with Dave Scott's Pointless-derived wm?
4.  Audible button clicks
5.  Integration with the iPAQ buttons
6.  Display of fscrib symbols used to launch same command
"""

import gtk
try: import GdkImlib
except: pass

import os
import sys
import string

ICON_SIZE = 16
def BUTTON_SIZE(): return ICON_SIZE+6
_icon_dirs = []

class RawWindow(gtk.GtkWindow):
    # - Base class for Launcher windows

    def __init__(self, size=None):
        gtk.GtkWindow.__init__(self, gtk.WINDOW_POPUP)
        if size:
            apply(self.set_usize, size)

    set_pos = gtk.GtkWindow.set_uposition

    def width(self): return self.get_window().width
    def height(self): return self.get_window().height

    def place_relative(self, widget, flags):
        # Place a new RawWindow relative to an existing widget
        # - Flags: 4/0 = horizontal/vertical
        pos = widget.get_allocation()
        wd = widget.get_window().toplevel
        pos = (pos[0]+wd.x, pos[1]+wd.y,
               pos[0]+pos[2]+wd.x, pos[1]+pos[3]+wd.y)
        w, h, sw, sh=self.width(), self.height(), gtk.screen_width(), gtk.screen_height()
        if flags & 4:
            x=pos[0]-w
            y=(pos[1]+pos[3]-h)/2
            if x<0: x=pos[2]
        else:
            y=pos[1]-h
            x=(pos[0]+pos[2]-w)/2
            if y<0: y=pos[3]
        self.set_pos(max(0, min(x, sw-w)), max(0, min(y, sh-h)))

    def load_icon(self, icon, name):
        # - Load the icon for a named application
        for path in _icon_dirs:
            file = os.path.join(path, icon)
            try: return gtk.GtkPixmap(self, file)
            except:
                try:
                    i = GdkImlib.Image(file)
                    if i.render(): return i.make_pixmap()
                except: pass
        return self.text_icon(name)

    def text_icon(self, name):
        # - Create suitable text for an icon
        n=string.split(name)
        if len(n) > 1:
            txt=string.join(map(lambda x:x[0], n),"")
        else:
            if n[0][0] in string.uppercase:
                txt=n[0][0]
            else:
                txt=n[0]
        w=-2 # - Magic number.  Is it "correct"?
        for i in range(len(txt)):
            w = w + self.get_style().font.width(txt[i])
            if w > ICON_SIZE:
                return gtk.GtkLabel(txt[:i])
        return gtk.GtkLabel(txt)

class LaunchButton(RawWindow):

    def __init__(self):

        # - Load the users preferences
        self.LoadPrefs()
        RawWindow.__init__(self, (BUTTON_SIZE(), BUTTON_SIZE()))

        # - Create the button
        button = gtk.GtkButton()
        self.__arrow = gtk.GtkArrow()
        self.__arrow.set(gtk.ARROW_UP, gtk.SHADOW_OUT)
        button.add(self.__arrow)
        button.connect("clicked", self.OnToggleMenu)
        self.add(button)

        # - Show the window
        self.connect("destroy", self.HideMenu)
        self.show_all()

        # - Place it
        self.place()

        self.__menu = None

    def place(self):

        # - Select the location
        if self.__placement & 2:
            y = 0
        else:
            y = gtk.screen_height()-self.height()
        if self.__placement & 1:
            x = 0
        else:
            x = gtk.screen_width()-self.width()
        self.set_pos(x, y)
        
        # - Set the _WIN_LAYER property if possible
        # *** THIS SEGFAULTS STUFF IN PROPERTY_CHANGE()
##         atom = _gtk.gdk_atom_intern("_WIN_LAYER", 1)
##         if str(atom) == None:
##             sys.stderr.write("warning:Unable to set _WIN_LAYER\n")
##         else:
##             print "Located:", str(atom)
##             int_atom = _gtk.gdk_atom_intern("INTEGER", 1)
##             if str(int_atom) != None:
##                 print "New atom:", str(int_atom)
##                 self.get_window().property_change(atom,
##                                                   int_atom,
##                                                   32,
##                                                   GDK.PROP_MODE_REPLACE,
##                                                   (10L,)
##                                                   )

        # - *** Can't set _WIN_LAYER - use crappy timeout hooks.
        self.__timeout = gtk.timeout_add(2000, self.OnTimeoutRaise)
        self.connect("destroy", self.OnDestroyRaise)

    def OnTimeoutRaise(self):
        self.get_window()._raise()
        return 1

    def OnDestroyRaise(self, widget):
        gtk.timeout_remove(self.__timeout)

    def LoadPrefs(self):
        # - Read the preferences

        global _icon_dirs

        # - Defaults
        self.__placement = 0 # - Bottom-Right
        _icon_dirs = [".", "icons"] # - Local directory & "icons" subdir
        self.__menuitems = []

        # - Prefs files to load
        files = [os.path.join(os.environ["HOME"],
                              ".foal/foalrc"),
                 "/etc/foal/foalrc"]
        for name in files:
            try: self.ReadPrefs(open(name, "r"), [self.__menuitems])
            except:
                pass

    def ReadPrefs(self, file, menu):
        # - Read a prefs file.

        # This includes icon directory(s),
        # icon program name substitutions and list of programs
        # and menus!
        # Also includes where to place the launcher button.

        global ICON_SIZE
        global _icon_dirs

        try:
            while 1:
                line = file.readline()
                if not line:
                    break
                line = string.strip(line)
                if not line or line[0] == "#":
                    continue

                try:
                    i = string.index(line, "(")
                    func = string.strip(line[:i])
                    data = eval(line[i:])
                    if type(data) != type(()):
                        data = (data,)
                except:
                    continue

                if func == "version":
                    pass
                elif func == "placement":
                    self.__placement = data[0]
                elif func == "icondir":
                    _icon_dirs.extend(list(data))
                elif func == "iconsize":
                    ICON_SIZE = data[0]
                elif func == "startmenu":
                    m = LaunchMenuSubMenu(data[0], data[1], [])
                    menu[0].append(m)
                    menu.insert(0, m.items)
                elif func == "endmenu":
                    del menu[0]
                elif func == "menuitem":
                    menu[0].append(LaunchMenuItem(data[0], data[1], data[2]))
                elif func == "verifyitem":
                    submenu = [LaunchMenuItem(data[0]+":Yes", "yes_tick.xpm", data[2]),
                               LaunchMenuItem(data[0]+":NO", "no_cross.xpm", "")]
                    menu[0].append(LaunchMenuSubMenu(data[0],data[1],submenu))
                elif func == "include":
                    try: self.ReadPrefs(open(data[0], "r"), menu)
                    except: pass
        except:
            sys.stderr.write("warning:error reading Launcher prefs\n")

    def HideMenu(self, w=None):
        if self.__menu:
            self.__menu.destroy()
            self.__menu = None            
            self.__arrow.set(gtk.ARROW_UP, gtk.SHADOW_OUT)

    def OnToggleMenu(self, widget):
        if self.__menu:
            self.HideMenu()
        else:
            self.__arrow.set(gtk.ARROW_DOWN, gtk.SHADOW_IN)
            self.__menu = LaunchMenu(self.__menuitems, self, 0)

class LaunchTip(RawWindow):
    def __init__(self, tip, widget, pos):
        RawWindow.__init__(self)
        f=gtk.GtkFrame()
        f.add(gtk.GtkLabel(tip))
        self.add(f)
        self.show_all()
        self.place_relative(widget, pos)

class LaunchMenu(RawWindow):

    def __init__(self, menu, widget, pos_flags):
        # - Create the menu window
        if pos_flags & 4:
            RawWindow.__init__(self, (BUTTON_SIZE()*len(menu), BUTTON_SIZE()))
            box = gtk.GtkHButtonBox()
        else:
            RawWindow.__init__(self, (BUTTON_SIZE(), BUTTON_SIZE()*len(menu)))
            box = gtk.GtkVButtonBox()
        box.set_child_ipadding(0, 0)
        box.set_child_size(BUTTON_SIZE(),BUTTON_SIZE())
        box.set_spacing(0)

        # - Add an entry for each menu item
        for entry in menu:
            button = gtk.GtkButton()
            button.add(self.load_icon(entry.icon, entry.name))
            button.set_data("command", entry)
            button.connect("clicked", self.OnLaunchCommand)
            box.pack_start(button)
        self.add(box)

        self.__submenu = None
        self.__submenu_name = None

        # - Place menu
        self.show_all()
        self.place_relative(widget, pos_flags)
        self.__pos_flags = pos_flags

        # - Start up dynamic tooltips
        self.__tip_timer = gtk.timeout_add(2000,
                                           self.NextTip)
        self.__tip_index = 0
        self.__tip = None
        self.__tips = box.children()[:]
        self.connect("destroy", self.StopTip)
        self.connect("destroy", self.KillSubMenu)

    def NextTip(self):
        if self.__tip:
            self.__tip.destroy()
            self.__tip = None
        if self.__tip_index < len(self.__tips):
            i = self.__tip_index
            self.__tip=LaunchTip(self.__tips[i].get_data("command").name, self.__tips[i], self.__pos_flags ^ 4)
            self.__tip_index = i+1
            return 1
        self.__tip_timer = None
        return 0

    def StopTip(self, widget=None):
        if self.__tip_timer:
            gtk.timeout_remove(self.__tip_timer)
            self.__tip_timer = None
        if self.__tip:
            self.__tip.destroy()
            self.__tip = None

    def KillSubMenu(self, widget=None):
        if self.__submenu:
            self.__submenu.destroy()
            self.__submenu = None
            self.__submenu_name = None

    def OnLaunchCommand(self, button):
        self.StopTip()
        submenu = self.__submenu_name
        self.KillSubMenu()
        item = button.get_data("command")
        if isinstance(item, LaunchMenuItem) and item.command:
            command = item.command
            if command[:6] == "#foal.":
                Foal(command[6:])
                return
            try:
                # - Is this the best way to launch an app?
                os.system(command + "&")
            except:
                sys.stderr.write("warning:failed to Launch "+command+"\n")
        elif submenu != item.name:
            self.__submenu = LaunchMenu(item.items, button, self.__pos_flags ^ 4)
            self.__submenu_name = item.name

class LaunchMenuItem:
    def __init__(self, name, icon, program):
        self.name = name
        self.icon = icon
        self.command = program

class LaunchMenuSubMenu:
    def __init__(self, name, icon, items):
        self.name = name
        self.icon = icon
        self.items = items

class LaunchDock(RawWindow):
    def __init__(self, launcher):
        RawWindow.__init__(self,(gtk.screen_width()-BUTTON_SIZE(),BUTTON_SIZE()))
        box = gtk.GtkHBox()
        box.set_spacing(0)
        self.box = box
        self.add(box)
        self.show_all()
        self.place_relative(launcher, 4)
    def DockWindow(self, xid):
        s = gtk.GtkSocket()
        f=gtk.GtkFrame()
        f.set_shadow_type(gtk.SHADOW_IN)
        f.add(s)
        self.box.pack_end(f, 0, 0)
        f.show_all()
        s.connect("destroy", f.destroy)
        s.steal(xid)

launcher = None
dock = None

# *** HACK
def ProcessEvents():
    while gtk.events_pending():
        gtk.mainiteration()

def Foal(func):
    global launcher, dock
    for f in string.split(func, ","):
        if (f=="open") and not launcher:
            launcher = LaunchButton()
        elif (f=="close") and launcher:
            launcher.destroy()
            launcher = None
        elif f=="quit":
            Foal("dockclose,close")
            gtk.mainquit()
        elif (f=="dockopen") and launcher and not dock:
            ProcessEvents()
            dock = LaunchDock(launcher)
        elif (f=="dockclose") and dock:
            dock.destroy()
            dock = None

if __name__ == "__main__":
    # - Create & run the Launcher
    Foal("open")
    gtk.mainloop()

        
        
